@@ -3,7 +3,7 @@
 
 -- |
 -- Module      : Criterion.Report
--- Copyright   : (c) 2011 Bryan O'Sullivan
+-- Copyright   : (c) 2009-2014 Bryan O'Sullivan
 --
 -- License     : BSD-style
 -- Maintainer  : bos@serpentine.com
@@ -14,14 +14,14 @@
 
 module Criterion.Report
     (
-      Report(..)
-    , formatReport
+      formatReport
     , report
+    , tidyTails
     -- * Rendering helper functions
     , TemplateException(..)
     , loadTemplate
     , includeFile
-    , templateDir
+    , getTemplateDir
     , vector
     , vector2
     ) where
@@ -29,53 +29,56 @@ module Criterion.Report
 import Control.Exception (Exception, IOException, throwIO)
 import Control.Monad (mplus)
 import Control.Monad.IO.Class (MonadIO(liftIO))
-import Criterion.Analysis (Outliers(..), SampleAnalysis(..))
-import Criterion.Config (cfgReport, cfgTemplate, fromLJ)
-import Criterion.Monad (Criterion, getConfig)
+import Control.Monad.Reader (ask)
+import Criterion.Monad (Criterion)
+import Criterion.Types
+import Data.Aeson.Encode (encodeToTextBuilder)
+import Data.Aeson.Types (toJSON)
 import Data.Data (Data, Typeable)
-import Data.Monoid (Last(..))
+import Data.Foldable (forM_)
 import GHC.Generics (Generic)
 import Paths_criterion (getDataFileName)
-import Statistics.Sample.KernelDensity (kde)
-import Statistics.Types (Sample)
+import Statistics.Function (minMax)
 import System.Directory (doesFileExist)
-import System.FilePath ((</>), isPathSeparator)
-import System.IO.Unsafe (unsafePerformIO)
+import System.FilePath ((</>), (<.>), isPathSeparator)
 import Text.Hastache (MuType(..))
 import Text.Hastache.Context (mkGenericContext, mkStrContext, mkStrContextM)
 import qualified Control.Exception as E
 import qualified Data.Text as T
 import qualified Data.Text.IO as T
 import qualified Data.Text.Lazy as TL
+import qualified Data.Text.Lazy.Builder as TL
 import qualified Data.Text.Lazy.IO as TL
 import qualified Data.Vector.Generic as G
 import qualified Data.Vector.Unboxed as U
 import qualified Text.Hastache as H
 
-data Report = Report {
-      reportNumber :: Int
-    , reportName :: String
-    , reportTimes :: Sample
-    , reportAnalysis :: SampleAnalysis
-    , reportOutliers :: Outliers
-    } deriving (Eq, Read, Show, Typeable, Data, Generic)
+-- | Trim long flat tails from a KDE plot.
+tidyTails :: KDE -> KDE
+tidyTails KDE{..} = KDE { kdeType   = kdeType
+                        , kdeValues = G.slice front winSize kdeValues
+                        , kdePDF    = G.slice front winSize kdePDF
+                        }
+  where tiny     = uncurry subtract (minMax kdePDF) * 0.005
+        omitTiny = G.length . G.takeWhile ((<= tiny) . abs)
+        front    = omitTiny kdePDF
+        back     = omitTiny . G.reverse $ kdePDF
+        winSize  = G.length kdePDF - front - back
 
--- | The path to the template and other files used for generating
--- reports.
-templateDir :: FilePath
-templateDir = unsafePerformIO $ getDataFileName "templates"
-{-# NOINLINE templateDir #-}
+-- | Return the path to the template and other files used for
+-- generating reports.
+getTemplateDir :: IO FilePath
+getTemplateDir = getDataFileName "templates"
 
 -- | Write out a series of 'Report' values to a single file, if
 -- configured to do so.
 report :: [Report] -> Criterion ()
 report reports = do
-  cfg <- getConfig
-  case cfgReport cfg of
-    Last Nothing -> return ()
-    Last (Just name) -> liftIO $ do
-      tpl <- loadTemplate [".",templateDir] (fromLJ cfgTemplate cfg)
-      TL.writeFile name =<< formatReport reports tpl
+  Config{..} <- ask
+  forM_ reportFile $ \name -> liftIO $ do
+    td <- getTemplateDir
+    tpl <- loadTemplate [td,"."] template
+    TL.writeFile name =<< formatReport reports tpl
 
 -- | Format a series of 'Report' values using the given Hastache
 -- template.
@@ -83,23 +86,38 @@ formatReport :: [Report]
              -> T.Text    -- ^ Hastache template.
              -> IO TL.Text
 formatReport reports template = do
+  templates <- getTemplateDir
   let context "report"  = return $ MuList $ map inner reports
-      context "include" = return $ MuLambdaM $ includeFile [templateDir]
+      context "json"    = return $ MuVariable (encode reports)
+      context "include" = return $ MuLambdaM $ includeFile [templates]
       context _         = return $ MuNothing
-      inner Report{..} = mkStrContextM $ \nym ->
+      encode v = TL.toLazyText . encodeToTextBuilder . toJSON $ v
+      inner r@Report{..} = mkStrContextM $ \nym ->
                          case nym of
-                           "name"     -> return $ MuVariable reportName
+                           "name"     -> return . MuVariable . H.htmlEscape .
+                                         TL.pack $ reportName
+                           "json"     -> return $ MuVariable (encode r)
                            "number"   -> return $ MuVariable reportNumber
-                           "times"    -> return $ vector "x" reportTimes
-                           "kdetimes" -> return $ vector "x" kdeTimes
+                           "iters"    -> return $ vector "x" iters
+                           "times"    -> return $ vector "x" times
+                           "cycles"   -> return $ vector "x" cycles
+                           "kdetimes" -> return $ vector "x" kdeValues
                            "kdepdf"   -> return $ vector "x" kdePDF
-                           "kde"      -> return $ vector2 "time" "pdf" kdeTimes kdePDF
+                           "kde"      -> return $ vector2 "time" "pdf" kdeValues kdePDF
                            ('a':'n':_)-> mkGenericContext reportAnalysis $
                                          H.encodeStr nym
                            _          -> mkGenericContext reportOutliers $
                                          H.encodeStr nym
-          where (kdeTimes,kdePDF) = kde 128 reportTimes
-  H.hastacheStr H.defaultConfig template context
+          where [KDE{..}]   = reportKDEs
+                iters       = measure measIters reportMeasured
+                times       = measure measTime reportMeasured
+                cycles      = measure measCycles reportMeasured
+      config = H.defaultConfig {
+                 H.muEscapeFunc = H.emptyEscape
+               , H.muTemplateFileDir = Just templates
+               , H.muTemplateFileExt = Just ".tpl"
+               }
+  H.hastacheStr config template context
 
 -- | Render the elements of a vector.
 --
@@ -184,7 +202,7 @@ loadTemplate paths name
     | any isPathSeparator name = T.readFile name
     | otherwise                = go Nothing paths
   where go me (p:ps) = do
-          let cur = p </> name
+          let cur = p </> name <.> "tpl"
           x <- doesFileExist cur
           if x
             then T.readFile cur `E.catch` \e -> go (me `mplus` Just e) ps
